package com.netifera.platform.internal.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.netifera.platform.api.model.IEntity;
import com.netifera.platform.api.model.IShadowEntity;
import com.netifera.platform.api.model.ITreeBuilder;
import com.netifera.platform.api.model.ITreeBuilderProvider;
import com.netifera.platform.model.FolderEntity;
import com.netifera.platform.model.TreeStructureContext;

public class TreeBuilder implements ITreeBuilder {
	private IShadowEntity treeRoot;
	private final List<ITreeBuilderProvider> extensionProviders;
	private Map<Long,List<IShadowEntity>> shadowsMap;
	private Map<Long,Map<String,IShadowEntity>> foldersMap;
	
	public TreeBuilder(final List<ITreeBuilderProvider> providers) {
		extensionProviders = providers;
	}

	public synchronized void setRoot(final IEntity rootEntity) {
		shadowsMap = new HashMap<Long,List<IShadowEntity>>();
		foldersMap = new HashMap<Long,Map<String,IShadowEntity>>();
		treeRoot = handleRootEntity(rootEntity);
	}
	
	public synchronized IShadowEntity getRoot() {
		return treeRoot;
	}
	
	public synchronized Set<IShadowEntity> addEntity(IEntity entity) {
		//NOTE this might be wrong, we're adding the entity under a shadow of the realm, but there might be many shadows of the realm
//		IShadowEntity realmEntity = treeRoot.getStructureContext().searchEntityById(entity.getRealmId());
		IShadowEntity realmEntity = getShadow(entity.getRealmId());
		if(realmEntity == null) {
			System.err.println("Realm entity not found for entity: "  + entity);
			return Collections.emptySet();
		}
		
		if (getShadow(entity.getId()) != null) return Collections.emptySet();

		return handleEntity(entity, realmEntity);
	}
	
	private IShadowEntity handleRootEntity(IEntity entity) {
		for(ITreeBuilderProvider extension : extensionProviders) {
			if(extension.isHandled(entity)) {
				IShadowEntity shadow = extension.createRoot(entity);
				List<IShadowEntity> list = new ArrayList<IShadowEntity>();
				list.add(shadow);
				shadowsMap.put(entity.getId(), list);
				return shadow;
			}
		}
		return null;
	}
	
	private Set<IShadowEntity> handleEntity(IEntity entity, IShadowEntity realmEntity) {
		Set<IShadowEntity> changes = new HashSet<IShadowEntity>();

		for (String tag: entity.getTags()) {
			IShadowEntity folder = createFolder(tag, realmEntity, changes);
			addChildEntity(entity, folder, changes);
		}
	
		for (ITreeBuilderProvider provider: extensionProviders) {
			if (provider.isHandled(entity)) {
				if (provider.isRealmRoot(entity) && entity.getTags().size() == 0) {
					// if doesnt have tags, add it to the realm root
					addChildEntity(entity, realmEntity, changes);
				} else {
					for (IEntity parent: provider.getParents(entity))
						for (IShadowEntity parentShadow: getShadows(parent.getId()))
							addChildEntity(entity, parentShadow, changes);
				}
			}
		}

		return changes;
	}

	public synchronized Set<IShadowEntity> updateEntity(IEntity entity) {

		IShadowEntity shadow = getShadow(entity.getId());
		if (shadow == null) return Collections.emptySet();

		IShadowEntity realmEntity = getShadow(entity.getRealmId());
		if(realmEntity == null) {
			System.err.println("realm entity not found for entity: "  + entity);
			return Collections.emptySet();
		}

		Set<IShadowEntity> changes = new HashSet<IShadowEntity>();

		// if some tag was removed, make sure the entity is not included in the tag folder
		List<IShadowEntity> shadows = getShadows(entity.getId());
		for (IShadowEntity child: shadows.toArray(new IShadowEntity[shadows.size()])) {
			IShadowEntity parent = ((TreeStructureContext)child.getStructureContext()).getParent();
			if (parent instanceof FolderEntity) {
				FolderEntity folder = (FolderEntity) parent;
				if (!entity.getTags().contains(folder.getTag())) {
					removeChildEntity(entity, parent, changes);
				}
			}
		}

		// make sure the entity is included in a folder for each of its tags
		if (entity.getTags().size() > 0) {
			for (String tag: entity.getTags()) {
				IShadowEntity folder = createFolder(tag, realmEntity, changes);
				if (!((TreeStructureContext)folder.getStructureContext()).hasChild(entity))
					copySubtree(shadow, folder, changes);
			}
			removeChildEntity(entity, realmEntity, changes);
		}

		changes.addAll(getShadows(entity.getId()));

		if (getShadow(entity.getId()) == null) {
			changes.addAll(handleEntity(entity, realmEntity));
		}
				
		return changes;
	}

	public synchronized Set<IShadowEntity> removeEntity(IEntity entity) {
		Set<IShadowEntity> changes = new HashSet<IShadowEntity>();
		Set<IShadowEntity> parents = new HashSet<IShadowEntity>();

		for (IShadowEntity shadow: getShadows(entity.getId())) {
			changes.add(shadow);
			TreeStructureContext context = (TreeStructureContext) shadow.getStructureContext();
			if (context.getParent() != null)
				parents.add(context.getParent());
		}
		
		for (IShadowEntity parent: parents)
			removeChildEntity(entity, parent, changes);

		return changes;
	}
	
	private List<IShadowEntity> getShadows(long id) {
		List<IShadowEntity> shadows = shadowsMap.get(id);
		if (shadows != null)
			return shadows;
		return Collections.emptyList();
	}

	private IShadowEntity getShadow(long id) {
		List<IShadowEntity> shadows = shadowsMap.get(id);
		if (shadows != null && shadows.size() > 0)
			return shadows.get(0);
		return null;
	}

	private IShadowEntity createFolder(String tag, IShadowEntity realmEntity, Set<IShadowEntity> changes) {
		Map<String,IShadowEntity> folderMap2 = foldersMap.get(realmEntity.getId());
		if (folderMap2 == null) {
			folderMap2 = new HashMap<String,IShadowEntity>();
			foldersMap.put(realmEntity.getId(), folderMap2);
		}
		IShadowEntity shadow = folderMap2.get(tag);
		if (shadow == null) {
			shadow = addChildEntity(new FolderEntity(realmEntity.getId(), null, tag), realmEntity, changes);
			folderMap2.put(tag, shadow);
		}
		return shadow;
	}
	
	private IShadowEntity addChildEntity(IEntity entity, IShadowEntity parent, Set<IShadowEntity> changes) {
		IShadowEntity shadow = ((TreeStructureContext) parent.getStructureContext()).addChild(entity);
		if (shadow != null) {
			if (!(entity instanceof FolderEntity)) {
				List<IShadowEntity> shadows = shadowsMap.get(entity.getId());
				if (shadows == null) {
					shadows = new ArrayList<IShadowEntity>();
					shadowsMap.put(entity.getId(), shadows);
				}
				shadows.add(shadow);
			}
			if (changes != null)
				changes.add(parent);
		}
		return shadow;
	}

	private IShadowEntity removeChildEntity(IEntity entity, IShadowEntity parent, Set<IShadowEntity> changes) {
		TreeStructureContext tsc = (TreeStructureContext)parent.getStructureContext();
		IShadowEntity shadow = tsc.removeChild(entity);
		if (shadow != null) {
			if (!(entity instanceof FolderEntity)) {
				List<IShadowEntity> shadows = shadowsMap.get(entity.getId());
				shadows.remove(shadow);
				if (shadows.size() == 0)
					shadowsMap.remove(entity.getId());
			}
			if (changes != null)
				changes.add(parent);
			
			// if the parent is a folder and becomes empty after removing this child, then remove the folder too
			if (parent instanceof FolderEntity && !tsc.hasChildren()) {
				Map<String,IShadowEntity> folderMap2 = foldersMap.get(parent.getRealmId());
				if (folderMap2 != null)
					folderMap2.remove(((FolderEntity)parent).getTag());
				IShadowEntity grandParent = tsc.getParent();
				if (grandParent != null)
					removeChildEntity(parent, grandParent, changes);
			}
		}
		return shadow;
	}
	
	private IShadowEntity copySubtree(IShadowEntity shadow, IShadowEntity parent, Set<IShadowEntity> changes) {
		IShadowEntity newShadow = addChildEntity(shadow.getRealEntity(), parent, changes);
		for (IShadowEntity child: ((TreeStructureContext)shadow.getStructureContext()).getChildren()) {
			copySubtree(child, newShadow, changes);
		}
		return newShadow;
	}
}
