package com.netifera.platform.net.internal.daemon.probe;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.netifera.platform.api.dispatcher.IProbeMessage;
import com.netifera.platform.api.dispatcher.MessengerException;
import com.netifera.platform.api.log.ILogger;
import com.netifera.platform.api.probe.IProbe;
import com.netifera.platform.dispatcher.StatusMessage;
import com.netifera.platform.net.daemon.sniffing.ISniffingDaemon;
import com.netifera.platform.net.daemon.sniffing.ISniffingModule;
import com.netifera.platform.net.pcap.ICaptureInterface;
import com.netifera.platform.net.sniffing.ICaptureFileInterface;
import com.netifera.platform.net.sniffing.util.ICaptureFileProgress;

public class RemoteSniffingDaemon implements ISniffingDaemon {
	private final IProbe probe;
	private final ILogger logger;
	private String messengerError;
	
	private ICaptureFileProgress progress;
	
	public RemoteSniffingDaemon(IProbe probe, ILogger logger) {
		this.probe = probe;
		this.logger = logger;
	}
	
	public ICaptureFileInterface createCaptureFileInterface(String path) {
		final CaptureFileValid valid = (CaptureFileValid) exchangeMessage(new CaptureFileValid(path));
		if(valid == null) {
			logger.warning("CaptureFileValid failed");
			return null;
		}
		return new CaptureFileStub(path, valid.isValid(), valid.getErrorMessage());
	}

	public void enableInterfaces(Collection<ICaptureInterface> interfaces) {
		final List<InterfaceRecord> interfaceRecords = new ArrayList<InterfaceRecord>();
		for(ICaptureInterface iface : interfaces) {
			interfaceRecords.add(new InterfaceRecord(iface.getName(), true, true));
		}
		if(!sendMessage(new SetInterfaceEnableState(interfaceRecords))) {
			logger.warning("Failed to set interface states: " + getLastError());
		}
	}

	public void enableModules(Set<ISniffingModule> enabledModuleSet) {
		final List<ModuleRecord> moduleRecords = new ArrayList<ModuleRecord>();
		for(ISniffingModule module : enabledModuleSet) {
			moduleRecords.add(new ModuleRecord(module.getName(), true));
		}
		if(!sendMessage(new SetModuleEnableState(moduleRecords))) {
			logger.warning("Failed to set module enable states: " + getLastError());
		}
		
	}

	public Collection<ICaptureInterface> getInterfaces() {
		final List<InterfaceRecord> interfaceRecords = getInterfaceRecords();
		if(interfaceRecords == null) {
			return Collections.emptyList();
		}
		return new ArrayList<ICaptureInterface>(interfaceRecords);
	}

	private List<InterfaceRecord> getInterfaceRecords() {
		final RequestInterfaceInformation response = (RequestInterfaceInformation) exchangeMessage(new RequestInterfaceInformation());
		if(response == null) {
			logger.warning("Failed to get capture interface information: " + getLastError());
			return null;
		}
		return response.getInterfaceRecords();
		
	}
	public Set<ISniffingModule> getModules() {
		final List<ModuleRecord> moduleRecords = getModuleRecords();
		if(moduleRecords == null) {
			return Collections.emptySet();
		}
		return new HashSet<ISniffingModule>(moduleRecords);
	}

	private List<ModuleRecord> getModuleRecords() {
		final RequestModuleInformation response = (RequestModuleInformation) exchangeMessage(new RequestModuleInformation());
		if(response == null) {
			logger.warning("Failed to get module information: " + getLastError());
			return null;
		}
		return response.getModuleRecords();
		
	}
	
	public boolean isEnabled(ICaptureInterface iface) {
		for(InterfaceRecord record : getInterfaceRecords()) {
			if(record.getName().equals(iface.getName()))
				return record.isEnabled();
		}
		logger.warning("Interface not found for name " + iface.getName());
		return false;
	}

	public boolean isEnabled(ISniffingModule module) {
		for(ModuleRecord record : getModuleRecords()) {
			if(record.getName().equals(module.getName())) {
				return record.isEnabled();
			}
		}
		logger.warning("Module not found for name " + module.getName());
		return false;
	}

	public void runCaptureFile(long spaceId, ICaptureFileInterface iface,
			ICaptureFileProgress progress) {
		this.progress = progress;
		
		if(!sendMessage(new RunCaptureFile(spaceId, iface.getPath()))) {
			logger.warning("Error running capture file " + getLastError());
		}
	}
	
	public void cancelCaptureFile() {
		if(!sendMessage(new CancelCaptureFile())) {
			logger.warning("Error cancelling capture file " + getLastError());
		}
	}

	public void setEnabled(ICaptureInterface iface, boolean enable) {
		final InterfaceRecord interfaceRecord = new InterfaceRecord(iface.getName(), iface.captureAvailable(), enable);
		if(!sendMessage(new SetInterfaceEnableState(interfaceRecord))) {
			logger.warning("Failed to enable interface " + getLastError());
		}		
	}

	public void setEnabled(ISniffingModule module, boolean enable) {
		final ModuleRecord moduleRecord = new ModuleRecord(module.getName(), enable);	
		if(!sendMessage(new SetModuleEnableState(moduleRecord))) {
			logger.warning("Failed to enable sniffing module " + getLastError());
		}	
	}

	public boolean isRunning() {
		SniffingDaemonStatus status = (SniffingDaemonStatus) exchangeMessage(new SniffingDaemonStatus());
		if(status == null) {
			logger.warning("Failed to get SniffingDaemonStatus");
			return false;
		}
		return status.isRunning;
	}

	public void start(long spaceId) {
		if(!sendMessage(new StartSniffingDaemon(spaceId))) {
			logger.warning("Failed to start sniffing daemon " + getLastError());
		}		
	}

	public void stop() {
		if(!sendMessage(new StopSniffingDaemon())) {
			logger.warning("Failed to stop sniffing daemon " + getLastError());
		}
		
	}
	
	public void captureFileProgress(CaptureFileProgress progressUpdate) {
		if(progressUpdate.isUpdate()) {
			progress.updateProgress(progressUpdate.getPercent(), progressUpdate.getCount());
		} else if(progressUpdate.isError()) {
			progress.error(progressUpdate.getErrorMessage(), null);
		} else {
			progress.finished();
		}
	}
	
	public void sniffingModuleOutput(String output) {
		logger.getManager().logRaw(output);
	}
	
	private String getLastError() {
		return messengerError;
	}
	private boolean sendMessage(IProbeMessage message) {
		try {
			probe.getMessenger().sendMessage(message);
			return true;
		} catch (MessengerException e) {
			messengerError = e.getMessage();
			return false;
		}
	}
	
	private IProbeMessage exchangeMessage(IProbeMessage message) {
		
		try {
			IProbeMessage response = probe.getMessenger().exchangeMessage(message);
			if(response instanceof StatusMessage) { 
				return null;
			} else {
				return response;
			}
		} catch (MessengerException e) {
			messengerError = e.getMessage();
			return null;
		}
	}

}
