package com.netifera.platform.net.sniffing;

import com.netifera.platform.net.pcap.ICaptureHeader;

/**
 * This interface provides access to information which is needed while
 * processing a captured packet.   It is passed as a parameter to the
 * callback methods of sniffing module implementations so that the
 * information is available during the handling and processing of
 * captured network information.
 * 
 * The information includes some data about the captured packet itself
 * as well as model related information (realms and spaces) which is required
 * when creating new entities.
 */
public interface IPacketContext {
	/**
	 * Return the capture header for this packet.
	 * 
	 * @return The capture header for this packet.
	 */
	ICaptureHeader getCaptureHeader();
	
	/**
	 * Returns the id value of the realm this packet was captured in.  Realms 
	 * are subsets of the entire data model where each entity is required to
	 * be uniquely identified.  As an example, two identical IP addresses can
	 * exist in the model but they must be in separate realms.  This allows 
	 * for representation of different private internal networks that have
	 * the same allocation of network ranges (ie: 192.168.x.x).
	 * 
	 * @see com.netifera.platform.net.sniffing.ISessionContext#getRealm()
	 * 
	 * @return The id of the current realm.
	 */
	long getRealm();
	
	/**
	 * Returns the id value of the space that tools and sniffing modules which 
	 * process this packet should create entities in.  In the user interface
	 * spaces are displayed as a collection of tabbed windows that the user
	 * can navigate.
	 * 
	 * @see com.netifera.platform.net.sniffing.IPacketContext#getSpaceId()
	 * 
	 * @return The id value for the space where new entities should be displayed.
	 */
	long getSpaceId();
	
	/**
	 * Changes the current realm for processing this packet.  While processing a packet,
	 * some tools and sniffing modules may discover information that places this packet
	 * into a different realm.  An example of when this might happen is while processing
	 * raw wireless packets captured in monitor mode.  Modules that process the link layer
	 * can identify that data frames belong to a specific wireless network and adjust the 
	 * realm so that higher layers create entities inside the realm for the wireless network.
	 * 
	 * @param realm The new realm id.
	 */
	void setRealm(long realm);

}
