package com.netifera.platform.ui.spaces.tree;

import java.util.Set;

import org.eclipse.jface.viewers.StructuredViewer;

import com.netifera.platform.api.events.IEvent;
import com.netifera.platform.api.events.IEventHandler;
import com.netifera.platform.api.model.IEntity;
import com.netifera.platform.api.model.IShadowEntity;
import com.netifera.platform.api.model.ISpace;
import com.netifera.platform.api.model.ISpaceContentChangeEvent;
import com.netifera.platform.api.model.ITreeBuilder;
import com.netifera.platform.model.TreeStructureContext;
import com.netifera.platform.ui.updater.StructuredViewerUpdater;

public class SpaceTreeUpdater {
	// private final static boolean EXPAND_PARENT = false;
	private final ITreeBuilder treeBuilder;
	private final StructuredViewer viewer;
	private final StructuredViewerUpdater treeViewer;
	private final IEventHandler spaceListener;
	private final ISpace space;

	SpaceTreeUpdater(ISpace space, StructuredViewer treeViewer) {
		if(space == null || treeViewer == null) {
			throw new IllegalArgumentException("space=" + space + ", viewer=" + treeViewer);
		}
		this.space = space;
		this.viewer = treeViewer;
		this.treeViewer = StructuredViewerUpdater.get(viewer);
		this.treeBuilder = space.getWorkspace().createTreeBuilder(space.getRootEntity());
		this.spaceListener = createSpaceListener();
		this.space.addChangeListenerAndPopulate(spaceListener);
	}
	
	public void dispose() {
		space.removeChangeListener(spaceListener);
	}
	
	public IShadowEntity getRootEntity() {
		return treeBuilder.getRoot();
	}
	
	public ISpace getSpace() {
		return space;
	}
	
	private IEventHandler createSpaceListener() {
		return new IEventHandler() {
			public void handleEvent(IEvent event) {
				if(event instanceof ISpaceContentChangeEvent) {
					handleSpaceChange((ISpaceContentChangeEvent)event);
				}
			}			
		};
	}
	
	private void handleSpaceChange(ISpaceContentChangeEvent event) {
		if(event.isCreationEvent()) {
			handleCreation(event.getEntity());
		} else if(event.isUpdateEvent()) {
			handleUpdate(event.getEntity());
		} else if(event.isRemovalEvent()) {
			handleRemoval(event.getEntity());
		}
	}

	private void handleCreation(final IEntity entity) {
		final Set<IShadowEntity> changes = treeBuilder.addEntity(entity);
		updateTreeForCreation(entity, changes);
	}
	
	private void updateTreeForCreation(IEntity entity, Set<IShadowEntity> changes) {
		if(treeViewer.getControl().isDisposed()) {
			return;
		}
		if(changes.isEmpty()) { //WTF
			treeViewer.refresh();
			return;
		}

		boolean refreshRoots = false;
		for(IShadowEntity shadow : changes) {
			IShadowEntity added = shadow.searchEntity(entity);
			if(added != null) {
				if (shadow.getRealEntity() != entity) {
					// if this shadow is a parent of entity that have just been created
					// (for example a newly created FolderEntity), update properly
					if ((shadow.getStructureContext() instanceof TreeStructureContext)
							&& ((TreeStructureContext)shadow.getStructureContext()).getChildren().size() <= 1)
						entityAdded(shadow);
				}
				entityAdded(added);
			} else {
				refreshRoots = true;
			}
		}
		
		if (refreshRoots)
			treeViewer.refresh();
	}
		
	private void entityAdded(IShadowEntity added) {
		if(!hasParent(added)) {
			treeViewer.refresh();
			return;
		}
		
		final IShadowEntity parent = getParent(added);
		treeViewer.refresh(parent);
/*		if(EXPAND_PARENT) {
			treeViewer.expandToLevel(parent, 1);
		}*/		
	}
	
	private boolean hasParent(IShadowEntity entity) {
		if(entity.getStructureContext() instanceof TreeStructureContext) {
			IShadowEntity parent = ((TreeStructureContext)entity.getStructureContext()).getParent();
			return (parent != null) && (parent != treeBuilder.getRoot());
		} else {
			return false;
		}
	}
		
	private IShadowEntity getParent(IShadowEntity entity) {
		return ((TreeStructureContext)entity.getStructureContext()).getParent();
	}
	
	private void handleUpdate(final IEntity entity) {
		if(!treeBuilderHasValidRoot()) {
			return;
		}
		
		final Set<IShadowEntity> changes = treeBuilder.updateEntity(entity);
		if(changes.isEmpty()) {
			return;
		}
				updateTreeForUpdate(entity, changes);
	}
	
	private boolean treeBuilderHasValidRoot() {
		return (treeBuilder.getRoot() != null && (treeBuilder.getRoot().getStructureContext() instanceof TreeStructureContext));
	}
	
	private void updateTreeForUpdate(IEntity entity, Set<IShadowEntity> changes) {
		if(treeViewer.getControl().isDisposed()) {
			return;
		}

		boolean refreshRoots = false;
		for(IShadowEntity shadow : changes) {
			if(shadow.getRealEntity() == entity) {
				treeViewer.update(shadow, null);
			} else {
				IShadowEntity found = shadow.searchEntity(entity);
				if(found != null && hasParent(found)) {
					IShadowEntity parent = getParent(found);
					treeViewer.refresh(parent);
					/*if(EXPAND_PARENT) {
						treeViewer.expandToLevel(parent, 1);
					}*/
				}
				if (!hasParent(shadow))
					refreshRoots = true;
				else
					treeViewer.refresh(shadow);
			}
		}
		if (refreshRoots)
			treeViewer.refresh();
	}
	
	private void handleRemoval(IEntity entity) {
		Set<IShadowEntity> changes = treeBuilder.removeEntity(entity);
		updateTreeForRemoval(entity, changes);
	}
	
	private void updateTreeForRemoval(IEntity entity, Set<IShadowEntity> changes) {
		if(treeViewer.getControl().isDisposed()) {
			return;
		}
		if(changes.isEmpty()) { //WTF
			treeViewer.refresh();
			return;
		}
		boolean refreshRoots = false;
		for(IShadowEntity shadow : changes) {
			treeViewer.refresh(shadow);
			if (shadow == treeBuilder.getRoot())
				refreshRoots = true;
		}
		if (refreshRoots)
			treeViewer.refresh();
	}
}

